/*
   to do:
   ------
   arrays
   on x goto/gosub
*/

#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <setjmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/***** typedefs *****/

typedef signed char schar;

struct numberline
{
  int number;
  schar *line;
  struct numberline *next;
};

struct numericvar
{
  double number;
  schar *name;
  struct numericvar *next;
};

struct stringvar
{
  schar *string;
  schar *name;
  struct stringvar *next;
};

typedef struct numberline nline;
typedef struct numericvar nvar;
typedef struct stringvar svar;

typedef struct
{
  char type;
  union
  {
    double number;
    schar *string;
  }value;
}rval;

enum
{
  NUMERIC,
  DYN_STRING,
  CONST_STRING
};

/***** end typedefs *****/

/***** ctype.h replacements *****/

int isactualspace(schar c)
{
  if(c<0) return 0;
  return isspace(c);
}

int isactualdigit(schar c)
{
  return c>='0' && c<='9';
}

int isactualalpha(char c)
{
  return c>='a' && c<='z' || c>='A' && c<='Z';
}

int isactualalnum(char c)
{
  return isactualalpha(c) || isactualdigit(c);
}

/***** end ctype.h replacements *****/

/***** tokens *****/

#define NTOKENS 51

enum
{
  AND=-1,
  OR=-2,
  GREATEREQUAL=-3,
  LESSEQUAL=-4,
  END=-5,
  RUN=-6,
  GOTO=-7,
  INPUT=-8,
  LIST=-9,
  PRINT=-10,
  SYSTEM=-11,
  IF=-12,
  THEN=-13,
  ELSE=-14,
  NEW=-15,
  FOR=-16,
  NEXT=-17,
  REM=-18,
  LOAD=-19,
  SAVE=-20,
  HELP=-21,
  GOSUB=-22,
  RETURN=-23,
  NOTEQUAL=-24,
  STEP=-25,
  TO=-26,
  INT=-27,
  SQR=-28,
  COS=-29,
  SIN=-30,
  TAN=-31,
  ATN=-32,
  ABS=-33,
  RND=-34,
  LET=-35,
  MOD=-36,
  SGN=-37,
  MIDSTRING=-38,
  LEN=-39,
  VAL=-40,
  STRSTRING=-41,
  INSTR=-42
};

typedef struct
{
  schar replace;
  schar *source;
} token;

token tokenlist[NTOKENS]=
{
  { AND,"and" },
  { OR,"or" },
  { GREATEREQUAL,">=" },
  { LESSEQUAL,"<=" },
  { END,"end" },
  { RUN,"run" },
  { GOTO,"goto" },
  { INPUT,"input" },
  { LIST,"list" },
  { PRINT,"print" },
  { PRINT,"?" },
  { SYSTEM,"system" },
  { IF,"if" },
  { THEN,"then" },
  { ELSE,"else" },
  { NEW,"new" },
  { FOR,"for" },
  { NEXT,"next" },
  { REM,"rem" },
  { SAVE,"save" },
  { LOAD,"load" },
  { HELP,"help" },
  { GOSUB,"gosub" },
  { RETURN,"return" },
  { NOTEQUAL,"<>" },
  { STEP,"step" },
  { TO,"to" },
  { INT,"int" },
  { SQR,"sqr" },
  { COS,"cos" },
  { SIN,"sin" },
  { TAN,"tan" },
  { ATN,"atn" },
  { ABS,"abs" },
  { RND,"rnd" },
  { LET,"let" },
  { MOD,"mod" },
  { SGN,"sgn" },
  { MIDSTRING,"mid$" },
  { LEN,"len" },
  { VAL,"val" },
  { STRSTRING,"str$" },
  { INSTR,"instr" },
  { '+',"+" },
  { '-',"-" },
  { '/',"/" },
  { '*',"*" },
  { '=',"=" },
  { '>',">" },
  { '<',"<" },
  { ':',":" }
};

schar *tokenize(schar *source, schar *dest)
{
  schar *d=dest;
  schar source2[256];
  schar *s2=source2;
  int i;
  int quot=0;

  strcpy(source2,source);

  for(i=0; i<strlen(source2); i++) source2[i]=tolower(source2[i]);

  while(*source)
  {
    if(!quot)
      if(isactualspace(*source))
      {
        if(d>dest && d[-1]!=' ') *d++=' ';
        while(isactualspace(*source)) source++,s2++;
      }
    if(*source=='"' && quot<2) quot=!quot;
    if(!quot)
      for(i=0; i<NTOKENS; i++)
      {
        if(!memcmp(s2,tokenlist[i].source,strlen(tokenlist[i].source)))
        {
          if(d>dest && d[-1]!=' ') *d++=' ';
          *d++=tokenlist[i].replace;
          *d++=' ';
          source+=strlen(tokenlist[i].source);
          s2+=strlen(tokenlist[i].source);
          if(tokenlist[i].replace==REM) quot=2;
          break;
        }
      }
    else i=NTOKENS;
    if(i==NTOKENS) *d++=*source++,s2++;
  }

  *d=0;

  return dest;
}

/***** end tokens *****/

/***** stacks *****/

#define GOSUBSTACKSIZE 1024
#define FORSTACKSIZE 256
#define IDENTLEN 16

nline *gosubstack[GOSUBSTACKSIZE];
schar *gosubcodestack[GOSUBSTACKSIZE];
int gosubstackpointer=0;

nline *forstack[FORSTACKSIZE];
schar *forcodestack[FORSTACKSIZE];
schar *forvardest[FORSTACKSIZE];
double forvarstep[FORSTACKSIZE];
schar *forvariable[FORSTACKSIZE];
int forstackpointer;

/***** end stacks *****/

/***** error handling *****/

char *IDENTTOOLONG="Identifier too long";
char *UNMATCHEDPAREN="Unmatched '('";
char *UNEXPECTEDDOT="More than one decimal point in that number";
char *LPARENEXPECTED="Expecting a '(' in that function call";
char *UNEXPECTEDCHAR="Unexpected character in expression";
char *UNEXPECTEDCHARPRINT="Unexpected character in print statement";
char *NOASSIGNMENT="Assignment expected in for statement";
char *SYNTAX="Huh?";
char *EXPECTEDTHEN="Expected then following condition in if statement";
char *BADGOTO="Goto attempted to nonexisting line";
char *EXPECTEDEOL="Expected end of line";
char *EMPTYEXPRESSION="Expression empty";
char *CANTLOAD="Unable to open file for loading";
char *CANTSAVE="Unable to open file for saving";
char *ILLEGALLOAD="Immediate command in file ignored";
char *DIVZERO="Divide by zero";
char *BADGOSUB="Gosub attempted to nonexisting line";
char *GOSUBOVERFLOW="Gosub stack overflow";
char *GOSUBUNDERFLOW="Return without gosub";
char *DOMAINERR="Domain error";
char *EXPECTEDTO="Expected to after assignment in for statement";
char *FOROVERFLOW="For stack overflow";
char *FORUNDERFLOW="Next without for";
char *FORNOTMATCHED="Next variable does not match that of latest for";
char *TYPEMISMATCH="Type mismatch";
char *COMMAEXPECTED="Expecting another parameter to that function";
char *COMMANOTEXPECTED="Not expecting another parameter to that funtion";
char *OUTOFRANGE="Value out of range";

schar *errorcode;

jmp_buf errorbuf;

void error(char *s)
{
  errorcode=s;
  longjmp(errorbuf,1);
}

/***** end error handling *****/

/***** globals n stuff *****/

char banner[]="Crusty Basic v0.82 by Jim Crawford - built on " __DATE__ " " __TIME__;

nline program;
nvar numvars;
svar stringvars;

nline *ip=0;
nline *oip=0;

schar *evalstring;

rval logicop();

/***** end globals n stuff *****/

/***** language variables *****/

void initmem()
{
  program.number=0;
  program.line="";
  program.next=0;
  numvars.number=0;
  numvars.name="";
  numvars.next=0;
  stringvars.string="";
  stringvars.name="";
  stringvars.next=0;
}

void resetmem()
{
  nvar *cv=numvars.next,*cn;
  svar *sv=stringvars.next,*sn;

  gosubstackpointer=0;
  forstackpointer=0;

  while(cv)
  {
    cn=cv->next;
    free(cv->name);
    free(cv);
    cv=cn;
  }
  numvars.next=0;
  while(sv)
  {
    sn=sv->next;
    free(sv->name);
    free(sv->string);
    free(sv);
    sv=sn;
  }
  stringvars.next=0;
}

rval fetchnum(schar *name)
{
  nvar *cnum=&numvars;
  rval ret;

  while(cnum->next && strcmp(name,cnum->name)) cnum=cnum->next;
  if(strcmp(name,cnum->name))
  {
    cnum->next=malloc(sizeof(nvar));
    cnum=cnum->next;
    cnum->name=malloc(strlen(name)+1);
    strcpy(cnum->name,name);
    cnum->next=0;
    cnum->number=0;
  }

  ret.type=NUMERIC;
  ret.value.number=cnum->number;

  return ret;
}

rval fetchstring(schar *name)
{
  svar *cstr=&stringvars;
  rval ret;

  while(cstr->next && strcmp(name,cstr->name)) cstr=cstr->next;
  if(strcmp(name,cstr->name))
  {
    cstr->next=malloc(sizeof(svar));
    cstr=cstr->next;
    cstr->name=malloc(strlen(name)+1);
    strcpy(cstr->name,name);
    cstr->next=0;
    cstr->string=malloc(1);
    *cstr->string=0;
  }

  ret.type=CONST_STRING;
  ret.value.string=cstr->string;

  return ret;
}

void storenum(schar *name, double num)
{
  nvar *cnum=&numvars;

  fetchnum(name);
  while(cnum->next && strcmp(name,cnum->name)) cnum=cnum->next;
  cnum->number=num;
}

void storestring(schar *name, schar *string)
{
  svar *cstr=&stringvars;

  fetchstring(name);
  while(cstr->next && strcmp(name,cstr->name)) cstr=cstr->next;
  cstr->string=realloc(cstr->string,strlen(string)+1);
  strcpy(cstr->string,string);
}

rval fetch(schar *name)
{
  if(*name=='$') return fetchstring(name);
  return fetchnum(name);
}

schar *identifier(schar *dest, schar *source)
{
  int i;
  schar *d=dest+1;

  for(i=0; i<IDENTLEN; i++)
    if(isactualalpha(*source)) *d++=*source++;
    else break;
  *d=0;

  if(*source=='$') *dest='$',source++;
  else strcpy(dest,dest+1);

  if(i==IDENTLEN) error(IDENTTOOLONG);

  return dest;
}

void addline(schar *s)
{
  int line=0;
  nline *cline=&program;
  nline *newline;

  while(isactualdigit(*s)) line*=10,line+=*s-'0',s++;
  while(isactualspace(*s)) s++;
  while(cline->next && line>=cline->next->number) cline=cline->next;
  if(cline->number!=line)
  {
    if(!*s) return;
    newline=malloc(sizeof(nline));
    newline->line=malloc(strlen(s)+1);
    newline->number=line;
    strcpy(newline->line,s);
    newline->next=cline->next;
    cline->next=newline;
  }
  else
  {
    if(!*s)
    {
      newline=&program;
      while(newline->next!=cline) newline=newline->next;
      newline->next=cline->next;
      free(cline->line);
      free(cline);
    }
    else
    {
      cline->line=realloc(cline->line,strlen(s)+1);
      cline->number=line;
      strcpy(cline->line,s);
    }
  }
}

/***** end language variables *****/

/***** expression evaluation *****/

#define num(x) ((x).value.number)
#define str(x) ((x).value.string)

schar *evalstring;

void freerval(rval a)
{
  if(a.type==DYN_STRING) free(str(a));
  a.type=CONST_STRING;
  str(a)="";
}

rval parsestring()
{
  rval a;
  int len;

  evalstring++;
  for(len=0; evalstring[len]!='"'; len++);

  a.type=DYN_STRING;
  str(a)=malloc(len+1);
  memcpy(str(a),evalstring,len);
  str(a)[len]=0;

  evalstring=strchr(evalstring,'"')+1;

  return a;
}

rval parsenumber()
{
  rval a;
  int b=0;

  a.type=NUMERIC;
  num(a)=0;
  while(isactualspace(*evalstring)) evalstring++;
  while(isactualdigit(*evalstring) || *evalstring=='.')
  {
    if(*evalstring=='.')
    {
      evalstring++;
      if(!b) b=1;
      else error(UNEXPECTEDDOT);
    }
    if(!b) num(a)*=10,num(a)+=*evalstring-'0';
    else b*=10,num(a)+=(double)(*evalstring-'0')/b;
    evalstring++;
  }

  return a;
}

int getparms(rval *parms, char *format)
{
  int i=0;

  while(isactualspace(*evalstring)) evalstring++;
  if(*evalstring!='(') error(LPARENEXPECTED);
  while(*format && *evalstring!=')')
  {
    evalstring++;
    parms[i]=logicop();
    switch(*format)
    {
      case 'n':
      case 'N': if(parms[i].type!=NUMERIC) error(TYPEMISMATCH);
                break;
      case 's':
      case 'S': if(parms[i].type==NUMERIC) error(TYPEMISMATCH);
                break;
    }
    while(isactualspace(*evalstring)) evalstring++;
    if(isupper(*++format) && *evalstring!=',') error(COMMAEXPECTED);
    i++;
  }
  if(*evalstring==',') error(COMMANOTEXPECTED);
  if(*evalstring!=')') error(UNMATCHEDPAREN);
  evalstring++;

  return i;
}

rval functionop()
{
  rval a;
  rval parms[3];

  a.type=NUMERIC;
  switch(*evalstring++)
  {
    case ABS:
      getparms(parms,"N"); a=parms[0];
      num(a)=abs(num(parms[0])); return a;
    case ATN:
      getparms(parms,"N");
      num(a)=atan(num(parms[0])); return a;
    case COS:
      getparms(parms,"N");
      num(a)=cos(num(parms[0])); return a;
    case MIDSTRING:
      if(getparms(parms,"SNn")>2)
      {
        if(strlen(str(parms[0]))<=(int)num(parms[1])+(int)num(parms[2])-1)
          error(OUTOFRANGE);
        if(num(parms[1])<0 || num(parms[2])<0)
          error(OUTOFRANGE);
        a.type=DYN_STRING;
        str(a)=malloc((int)num(parms[2])+1);
        memcpy(str(a),str(parms[0])+(int)num(parms[1])-1,(int)num(parms[2]));
        str(a)[(int)num(parms[2])]=0;
        freerval(parms[0]);
      }
      else
      {
        if(strlen(str(parms[0]))<=(int)num(parms[1])-1)
          error(OUTOFRANGE);
        if(num(parms[1])<0)
          error(OUTOFRANGE);
        a.type=DYN_STRING;
        str(a)=malloc(strlen(str(parms[0])+(int)num(parms[1])-1)+1);
        strcpy(str(a),str(parms[0])+(int)num(parms[1])-1);
        freerval(parms[0]);
      }
      return a;
    case STRSTRING:
      getparms(parms,"N");
      a.type=DYN_STRING;
      str(a)=malloc(256);
      sprintf(str(a),"%g",num(parms[0]));
      str(a)=realloc(str(a),strlen(str(a))+1);
      return a;
    case INSTR:
      getparms(parms,"SS");
      num(a)=(int)(strstr(str(parms[0]),str(parms[1]))-str(parms[0]))+1;
      freerval(parms[0]);
      freerval(parms[1]);
      return a;
    case SIN:
      getparms(parms,"N");
      num(a)=sin(num(parms[0])); return a;
    case TAN:
      getparms(parms,"N");
      num(a)=tan(num(parms[0])); return a;
    case SQR:
      getparms(parms,"N");
      num(a)=sqrt(num(parms[0])); return a;
    case INT:
      getparms(parms,"N");
      num(a)=(long)num(parms[0]); return a;
    case LEN:
      getparms(parms,"S");
      num(a)=strlen(str(parms[0]));
      freerval(parms[0]);
      return a;
    case VAL:
      getparms(parms,"S");
      num(a)=atol(str(parms[0]));
      freerval(parms[0]);
      return a;
    case RND:
      while(isspace(*evalstring)) evalstring++;
      if(*evalstring=='(') getparms(parms,"N");
      num(a)=(double)rand()/INT_MAX; return a;
    case SGN:
      getparms(parms,"N");
      num(a)=0;
      if(num(parms[0])>0) num(a)=1;
      if(num(parms[0])<0) num(a)=-1;
      return a;
    default: error(UNEXPECTEDCHAR);
  }
  return a;
}

rval dataop()
{
  rval a;
  schar name[IDENTLEN+1];

  while(isactualspace(*evalstring)) evalstring++;
  if(*evalstring=='(')
  {
    evalstring++;
    a=logicop();
    if(*evalstring!=')') error(UNMATCHEDPAREN);
    evalstring++;
    return a;
  }
  if(*evalstring=='"') return parsestring();
  if(isactualdigit(*evalstring) || *evalstring=='.') return parsenumber();
  if(isactualalpha(*evalstring))
  {
    identifier(name,evalstring);
    evalstring+=strlen(name);
    return fetch(name);
  }
  else a=functionop();

  return a;
}

rval unaryop()
{
  rval a;

  while(isactualspace(*evalstring)) evalstring++;
  if(*evalstring=='-')
  {
    evalstring++;
    a=dataop();
    if(a.type!=NUMERIC) error(TYPEMISMATCH);
    num(a)=-num(a);
    return a;
  }

  return dataop();
}

rval powop()
{
  rval a,b;

  a=unaryop();

  for(;;)
  {
    while(isactualspace(*evalstring)) evalstring++;
    if(*evalstring!='^') return a;
    if(a.type!=NUMERIC) error(TYPEMISMATCH);
    evalstring++;
    b=unaryop();
    if(b.type!=NUMERIC) error(TYPEMISMATCH);
    num(a)=pow(num(a),num(b));
  }
}

rval mulop()
{
  rval a,b;
  char c;

  a=powop();

  for(;;)
  {
    while(isactualspace(*evalstring)) evalstring++;
    if(*evalstring!='*' && *evalstring!='/' && *evalstring!=MOD) return a;
    if(a.type!=NUMERIC) error(TYPEMISMATCH);
    c=*evalstring++;
    b=powop();
    if(b.type!=NUMERIC) error(TYPEMISMATCH);
    switch(c)
    {
      case '*': num(a)*=num(b); break;
      case '/': num(a)/=num(b); break;
      case MOD: num(a)=(long)num(a)%(long)num(b); break;
    }
  }
}

rval addop()
{
  rval a,b;
  schar *s;

  a=mulop();

  for(;;)
  {
    while(isactualspace(*evalstring)) evalstring++;
    if(*evalstring!='+' && *evalstring!='-') return a;
    switch(*evalstring++)
    {
      case '+': if(a.type==NUMERIC)
                {
                  b=mulop();
                  if(b.type!=NUMERIC) error(TYPEMISMATCH);
                  num(a)+=num(b);
                  break;
                }
                b=mulop();
                if(b.type==NUMERIC) error(TYPEMISMATCH);
                if(a.type==DYN_STRING)
                {
                  str(a)=realloc(str(a),strlen(str(a))+strlen(str(b))+1);
                  strcat(str(a),str(b));
                  freerval(b);
                  break;
                }
                s=malloc(strlen(str(a))+strlen(str(b))+1);
                strcpy(s,str(a));
                strcat(s,str(b));
                freerval(b);
                str(a)=s;
                a.type=DYN_STRING;
                break;
      case '-': if(a.type!=NUMERIC) error(TYPEMISMATCH);
                b=mulop();
                if(b.type!=NUMERIC) error(TYPEMISMATCH);
                num(a)-=num(b); break;
    }
  }
}

rval equop()
{
  rval a,b,d;
  schar c;

  a=addop();

  for(;;)
  {
    while(isactualspace(*evalstring)) evalstring++;
    c=*evalstring;

    if(c=='=' || c=='<' || c=='>' ||
       c==NOTEQUAL || c==GREATEREQUAL || c==LESSEQUAL)
    {
      evalstring++;
      b=addop();
      if((a.type==NUMERIC && b.type!=NUMERIC) ||
         (a.type!=NUMERIC && b.type==NUMERIC)) error(TYPEMISMATCH);
    }
    else return a;

    d=a;
    switch(c)
    {
      case '=': if(a.type==NUMERIC) num(a)=num(a)==num(b);
                else num(a)=!strcmp(str(a),str(b));
                break;
      case '>': if(a.type==NUMERIC) num(a)=num(a)>num(b);
                else num(a)=strcmp(str(a),str(b))>0;
                break;
      case '<': if(a.type==NUMERIC) num(a)=num(a)<num(b);
                else num(a)=strcmp(str(a),str(b))<0;
                break;
      case GREATEREQUAL:
                if(a.type==NUMERIC) num(a)=num(a)>=num(b);
                else num(a)=strcmp(str(a),str(b))>=0;
                break;
      case LESSEQUAL:
                if(a.type==NUMERIC) num(a)=num(a)<=num(b);
                else num(a)=strcmp(str(a),str(b))<=0;
                break;
      case NOTEQUAL:
                if(a.type==NUMERIC) num(a)=num(a)!=num(b);
                else num(a)=strcmp(str(a),str(b))!=0;
                break;
    }
    freerval(d);
    freerval(b);
    a.type=NUMERIC;
  }
}

rval logicop()
{
  rval a;

  a=equop();
  for(;;)
  {
    while(isactualspace(*evalstring)) evalstring++;
    switch(*evalstring)
    {
      case AND: if(a.type!=NUMERIC) error(TYPEMISMATCH);
                evalstring++;
                num(a)=num(equop())&&num(a);
                break;
      case OR:  if(a.type!=NUMERIC) error(TYPEMISMATCH);
                evalstring++;
                num(a)=num(equop())||num(a);
                break;
      default: return a;
    }
  }
}

double evaluate(schar *s)
{
  rval a;
  evalstring=s;
  a=logicop();
  if(a.type!=NUMERIC) error(TYPEMISMATCH);
  return num(a);
}

rval stringevaluate(schar *s)
{
  rval a;
  evalstring=s;
  a=logicop();
  if(a.type==NUMERIC) error(TYPEMISMATCH);
  return a;
}

/***** language functions *****/

schar *listfunction(schar *s)
{
  nline *cline=&program;
  schar *line;
  int i;
  int start=0,end=INT_MAX;

  while(isactualspace(*s)) s++;
  if(*s)
    if(*s=='-')
    {
      s++;
      while(isactualspace(*s)) s++;
      end=0;
      while(isactualdigit(*s))
        end*=10,end+=*s-'0',s++;
    }
    else
    {
      while(isactualdigit(*s))
        start*=10,start+=*s-'0',s++;
      while(isactualspace(*s)) s++;
      if(*s=='-')
      {
        s++;
        while(isactualspace(*s)) s++;
        end=0;
        while(isactualdigit(*s))
          end*=10,end+=*s-'0',s++;
        if(!end) end=INT_MAX;
      }
      else end=start;
    }

  while(cline->next)
  {
    cline=cline->next;
    if(cline->number>=start && cline->number<=end)
    {
      printf("%d ",cline->number);
      for(line=cline->line; *line; line++)
        if(*line>=0) putchar(*line);
        else
          for(i=0; i<NTOKENS; i++)
            if(tokenlist[i].replace==*line)
            {
              printf("%s",tokenlist[i].source);
              break;
            }

      putchar('\n');
    }
  }

  return s;
}

schar *printfunction(schar *s)
{
  int keep=0;
  rval a;

  while(isactualspace(*s)) s++;
  while(*s && *s!=':' && *s!=ELSE)
  {
    if(*s==';') keep=1,s++;
    else if(*s==',') 
         {
           putchar('\t');
           s++,keep=1;
         }
    else
    {
      evalstring=s;
      a=logicop();
      s=evalstring;
      if(a.type==NUMERIC) printf("%g",num(a));
      else printf("%s",str(a));
      if(a.type==DYN_STRING) freerval(a);
      keep=0;
    }
    while(isactualspace(*s)) s++;
  }
  if(!keep) putchar('\n');

  return s;
}

schar *assignfunction(schar *s)
{
  schar name[IDENTLEN+1];
  double n;
  rval string;

  while(isactualspace(*s)) s++;
  if(!*identifier(name,s)) return 0;
  s+=strlen(name);
  while(isactualspace(*s)) s++;
  if(*s!='=') error(SYNTAX);
  s++;
  while(isactualspace(*s)) s++;
  if(*name!='$') n=evaluate(s);
  else string=stringevaluate(s);
  s=evalstring;
  if(*name!='$') storenum(name,n);
  else 
  {
    storestring(name,str(string));
    freerval(string);
  }

  return s;
}

schar *inputfunction(schar *s)
{
  schar name[IDENTLEN+1];
  schar input[256];
  double d;

  while(isactualspace(*s)) s++;
  if(!*identifier(name,s)) return 0;
  s+=strlen(name);
  gets(input);
  if(*name!='$')
  {
    d=atof(input);
    storenum(name,d);
  }
  else storestring(name,input);

  return s;
}

schar *newfunction(char *s)
{
  nline *cl=program.next,*cn;

  while(cl)
  {
    cn=cl->next;
    free(cl->line);
    free(cl);
    cl=cn;
  }
  program.next=0;
  ip=0;

  return "";
}

schar *loadfunction(schar *s)
{
  FILE *f;
  schar input[256];
  schar tokenized[256];
  rval filename;

  while(isactualspace(*s)) s++;
  filename=stringevaluate(s);
  f=fopen(str(filename),"r");
  freerval(filename);
  if(!f) error(CANTLOAD);

  newfunction("");
  do
  {
    fgets(input,256,f);
    *strchr(input,'\n')=0;
    s=tokenize(input,tokenized);
    while(isactualspace(*s)) s++;
    if(!*s) continue;
    if(!isactualdigit(*s)) printf("Immediate in load ignored\n");
    else addline(s);
  }while(!feof(f));
  fclose(f);

  return evalstring;
}

schar *savefunction(schar *s)
{
  nline *cline=&program;
  schar *line;
  int i;
  FILE *f;
  rval filename;

  while(isactualspace(*s)) s++;
  filename=stringevaluate(s);
  f=fopen(str(filename),"w");
  freerval(filename);
  if(!f) error(CANTSAVE);

  while(cline->next)
  {
    cline=cline->next;
    fprintf(f,"%d ",cline->number);
    for(line=cline->line; *line; line++)
      if(*line>=0) fprintf(f,"%c",*line);
      else
        for(i=0; i<NTOKENS; i++)
          if(tokenlist[i].replace==*line)
          {
            fprintf(f,"%s",tokenlist[i].source);
            break;
          }
    fprintf(f,"\n");
  }
  fclose(f);

  return evalstring;
}

schar *helpfunction(char *s)
{
  printf("%s",banner);
  printf("\nfunctionality:"
         "\n<var>=<expression>"
         "\nend run system rem help list"
         "\nint cos sqr sin tan atn rnd abs sgn len val mid$ str$ instr"
         "\ngoto <expression>"
         "\ngosub <expression> / return"
         "\ninput <var>"
         "\nprint [{ <expression>|','|';' }]"
         "\nif <expression> then <statement> else <statement>"
         "\nfor <var>=<expression> to <expression> / next [<var>]"
         "\nload <string>"
         "\nsave <string>"
         "\n\npoints to remember:"
         "\nvars are case sensitive"
         "\noperator precedence: ( ) | ^ | * / | + - | <> >= <= = | and or "
         "\nif/else statements do not nest!  each if matches the closest else"
         "\nline number printed in error occuring right after a next/return is wrong\n");

  return "";
}

/***** end language functions *****/

/***** flow control functions *****/

schar *iffunction(schar *s)
{
  int condition;
  schar *elsep;

  while(isactualspace(*s)) s++;
  condition=(int)evaluate(s);
  s=evalstring;
  while(isactualspace(*s)) s++;

  if(*s!=THEN) error(EXPECTEDTHEN);
  if(condition) return s;

  elsep=strchr(s,(char)ELSE);
  if(!elsep) return "";

  return elsep;
}

schar *runfunction(schar *s)
{
  resetmem();
  ip=program.next;

  return s;
}

schar *endfunction()
{
  ip=0;

  return "";
}

schar *gotofunction(schar *s)
{
  int line;
  nline *cline=&program;

  line=(int)evaluate(s);

  while(cline->next)
  {
    cline=cline->next;
    if(cline->number==line)
    {
      ip=cline;
      return "";
    }
  }

  ip=0;
  error(BADGOTO);

  return "";
}

schar *gosubfunction(schar *s)
{
  int line;
  nline *cline=&program;

  if(gosubstackpointer>=GOSUBSTACKSIZE) error(GOSUBOVERFLOW);

  line=(int)evaluate(s);
  s=evalstring;

  while(cline->next)
  {
    cline=cline->next;
    if(cline->number==line)
    {
      gosubstack[gosubstackpointer]=ip;
      gosubcodestack[gosubstackpointer++]=s;
      ip=cline;

      return "";
    }
  }

  ip=0;
  error(BADGOSUB);

  return "";
}

schar *returnfunction(schar *s)
{
  if(--gosubstackpointer<0)
  {
    gosubstackpointer=0;
    error(GOSUBUNDERFLOW);
  }

  ip=gosubstack[gosubstackpointer];
  return gosubcodestack[gosubstackpointer];
}

schar *forfunction(schar *s)
{
  schar variable[IDENTLEN+1];
  double start;

  if(forstackpointer>=FORSTACKSIZE)
  {
    forstackpointer=0;
    error(FOROVERFLOW);
  }

  while(isactualspace(*s)) s++;

  if(!identifier(variable,s)) return "";
  if(*variable=='$') error(TYPEMISMATCH);
  forvariable[forstackpointer]=s;
  s+=strlen(variable);

  while(isactualspace(*s)) s++;
  if(*s!='=') error(NOASSIGNMENT);
  start=evaluate(++s);
  s=evalstring;

  storenum(variable,start);

  while(isactualspace(*s)) s++;

  if(*s!=TO) error(EXPECTEDTO);

  s++;
  while(isactualspace(*s)) s++;

  forstack[forstackpointer]=ip;
  forvardest[forstackpointer]=s;

  evaluate(s);
  s=evalstring;

  while(isactualspace(*s)) s++;

  if(*s!=STEP)
  {
    forcodestack[forstackpointer]=s;
    forvarstep[forstackpointer++]=1;

    return s;
  }

  s++;
  while(isactualspace(*s)) s++;
  forvarstep[forstackpointer]=evaluate(s);
  forcodestack[forstackpointer++]=evalstring;

  return evalstring;
}

schar *nextfunction(schar *s)
{
  double var,compare;
  schar name[IDENTLEN+1],name2[IDENTLEN+1];

  while(isactualspace(*s)) s++;

  if(isactualalpha(*s))
  {
    if(!identifier(name2,s)) return "";
    s+=strlen(name2);
  }
  else *name2=0;

  if(--forstackpointer<0)
  {
    gosubstackpointer=0;
    error(FORUNDERFLOW);
  }

  var=num(fetch(identifier(name,forvariable[forstackpointer])));

  if(strcmp(name,name2) && *name) error(FORNOTMATCHED);

  var=num(fetch(identifier(name,forvariable[forstackpointer])));
  var+=forvarstep[forstackpointer];
  compare=evaluate(forvardest[forstackpointer]);
  storenum(name,var);
  if((forvarstep[forstackpointer]>0 && var>compare) ||
     (forvarstep[forstackpointer]<0 && var<compare))
    return s;

  ip=forstack[forstackpointer];
  return forcodestack[forstackpointer++];
}

/***** end flow control functions *****/

/***** main loop *****/

int execute(schar *s)
{
  int skipflag;

  do
  {
    while(isactualspace(*s)) s++;
    skipflag=0;
    switch(*s)
    {
      case END: s=endfunction(); break;
      case GOTO: s=gotofunction(s+1); break;
      case INPUT: s=inputfunction(s+1); break;
      case LIST: s=listfunction(s+1); break;
      case PRINT: s=printfunction(s+1); break;
      case RUN: s=runfunction(s+1); break;
      case IF: s=iffunction(s+1);
               if(*s==THEN || *s==ELSE)
               {
                 s++,skipflag=1;
                 while(isactualspace(*s)) s++;
                 if(isactualdigit(*s)) s=gotofunction(s),skipflag=0;
               }
               
               break;
      case SYSTEM: return 255;
      case LOAD: s=loadfunction(s+1); break;
      case SAVE: s=savefunction(s+1); break;
      case REM: return 0;
      case HELP: s=helpfunction(s+1); break;
      case GOSUB: s=gosubfunction(s+1); break;
      case RETURN: s=returnfunction(s+1); break;
      case NEW: s=newfunction(s+1); break;
      case FOR: s=forfunction(s+1); break;
      case NEXT: s=nextfunction(s+1); break;
      case LET: s++;
      default: s=assignfunction(s); break;
    }
    while(isactualspace(*s)) s++;
  } while((*s==':' && *++s) || skipflag);

  while(isactualspace(*s)) s++;
  if(*s && *s!=ELSE && !errorcode) error(EXPECTEDEOL);

  return 0;
}

void mainloop()
{
  schar input[256];
  schar tokenized[256];
  schar *s;
  int q;

  do
  {
    if(setjmp(errorbuf))
    {
      printf("\n*** ");
      if(oip)
        printf("line %d: ",oip->number);

      printf("%s\n",errorcode);
      ip=0;
    }

    if(!ip)
    {
      do
      {
        oip=0;
/*        printf("\n%u>",coreleft());*/
        printf("\n>");
        gets(input);
        s=tokenize(input,tokenized);
        while(isactualspace(*s)) s++;
      } while(!*s);
      if(isactualdigit(*s)) addline(s);
      else q=execute(s);
    }
    else
    {
      s=ip->line;
      oip=ip;
      ip=ip->next;
      while(isactualspace(*s)) s++;
      q=execute(s);
    }
  }while(q!=255);
}

/***** end main loop *****/

main()
{
  printf("%s\nHELP for help (not very helpful unless you already know basic)"
         "\n",banner);
  initmem();
  mainloop();

  return 0;
}

